import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;

/**
 * Fenetre qui affiche le jeu, le hero et les monstres et
 * les evenements claviers.
 *
 * @version    1.0 17 Nov 1999
 * @author     Stephane Caillette & Thomas Salamito
 */ 
public class FenetreJeu extends Canvas implements Runnable, KeyListener {
 /* Classe dependant des evenements claviers et affichant le jeu. */
  // attribut
  /** La fenetre de notre programme. */
  private FenetrePrincipale fp;
  /** Les evenements claviers sont pris oui ou non en compte. */
  private boolean touche;
  /** Variable nous precisant si le jeu est parti ou non. */
  private boolean jeu;
  /** Variable de type Level. */
  private Level level;
  /** hauteur et largeur de la fenetre de jeu. */
  private int largeur, hauteur;
  /** Nombre de monstres presents dans le jeu. */
  private int nbMonstre;
  /** Tableau gerant les differents monstres. */
  private Monstre[] tableauDeMonstre;
  /** Notre heros. */
  private Heros heros;
  /** Image sauvegardee en memoire dans db pour ne pas tout redessiner. */
  private Image db;
  /** Contexte graphique de db. */
  private Graphics dbGraphics;
  /** Le thread qui anime le jeu. */
  private Thread thread;
  /** Image que l'on affiche au demarrage. */ 
  private String imageAAfficher;

  // constructeur 
  /**
   * Construit la zone de jeu.
   * @param fp La fenetre dans laquelle se deroule le jeu.
   */ 
  public FenetreJeu(FenetrePrincipale fp) {
    largeur = Sprite.getTaille() * Level.getNbDalleX();
    hauteur = Sprite.getTaille() * Level.getNbDalleY();
    setBackground(Color.black);

    this.fp = fp;

    level = new Level(this);
    thread = new Thread(this);

    jeu = false;
    imageAAfficher = "loderunner.jpg";
    // on gere les touches du clavier
    addKeyListener(this);

    this.initialiseToi();
  }

  // methode
  /**
   * Elle est lancee au demarrage.
   * Elle charge les dalles, les echelles,
   * les cadeaux et cree le heros et les monstres. 
   * On affecte l'observer FenetrePrincipale
   * a l'observable qui est notre hero.
   */ 
  public void initialiseToi() {

    nbMonstre = (level.getNiveau() / 2) + 1;

    level.charge();

    heros = new Heros(Level.getPosXInitHeros() * Sprite.getTaille(), 
		      Level.getPosYInitHeros() * Sprite.getTaille(), "mario_d",this);
    heros.addObserver(fp);
    
    tableauDeMonstre = new Monstre[3];
    for(int i = 0 ; i < 3 ; i++) {
      tableauDeMonstre[i] = new Monstre(Level.getPosXInitMonstre(i + 1) * Sprite.getTaille(), 
					Level.getPosYInitMonstre(i + 1) * Sprite.getTaille(), "monstre",this);
    }
  }

  /**
   * Elle est appellee lorsque le heros a fini 
   * le tableau.
   */
  public void niveauSuivant() {
    level.suivant();
    this.initialiseToi();
  }

  /**
   * Elle est appellee quand on a fini tous
   * les niveaux. 
   */
  public void finDuJeu() {
    thread.stop();
    jeu = false;
    imageAAfficher = "endgame.jpg";
    repaint();
    this.initialiseToi();
    touche = true;
  }

  /**
   * Elle est appellee quand on a perdu. 
   */
  public void gameOver() {
    thread.stop();
    jeu = false;
    imageAAfficher = "gameover.jpg";
    repaint();
    level.reset();
    this.initialiseToi();
    touche = true;
  }

  /**
   * Elle est appelle quand on veut arreter
   * momentanement le jeu. Si l'on veut repartir
   * on doit l'appeler une deuxieme fois.
   */
  public void pause() {
    if(thread.isAlive())
      thread.stop();
    else {
      thread = new Thread(this);
      thread.start();
    }
  }

  /**
   * Elle est definie pour que les actions 
   * des touches fonctionnent sous Windows.
   * @return true Necessaire pour que les touches fonctionnent sous Windows.
   */
  public boolean isFocusTraversable() {
    return true;
  }
    
  /** 
   * Elle sauve le fichier courant
   * pour le recharger ulterieurement.
   * @param fichierDeSauvegarde Le nom du fichier sauvegardee.
   */
  public void sauvegardeToi(String fichierDeSauvegarde) {
    if(jeu) {

      try {
	thread.sleep(5000); // stabilise le jeu
      } catch(InterruptedException e) {
	System.err.println("CLASS FenetreJeu : sleep erreur !");
	e.printStackTrace();
      }

      try {
	PrintWriter pw = new PrintWriter(new BufferedWriter(new FileWriter(fichierDeSauvegarde)));
	level.sauvegardeToi(pw);
	
	heros.sauvegardeToi(pw);
	
	for(int i = 0 ; i < nbMonstre ; i++) {
	  tableauDeMonstre[i].sauvegardeToi(pw);
	}

	pw.close();
      } catch(IOException ex) {
	System.err.println("CLASS FenetreJeu : echec lors de l'ouverture du fichier : " + fichierDeSauvegarde);
	ex.printStackTrace();
      }
    }
  }

  /** 
   * Elle charge le fichier courant.
   * @param fichierACharger c'est le nom du fichier charge.
   */
  public void chargeToi(String fichierACharger) {
    try {
      BufferedReader br = new BufferedReader(new FileReader(fichierACharger));
      StringTokenizer st = new StringTokenizer(br.readLine(), "|");
      level.chargeToi(st);

      heros.chargeToi(st);

      nbMonstre = (level.getNiveau() / 2) + 1;

      for(int i = 0 ; i < nbMonstre ; i++) {
	tableauDeMonstre[i].chargeToi(st);
      }
      
      br.close();
    } catch(IOException ex) {
      System.err.println("CLASS FenetreJeu : echec lors de l'ouverture du fichier : " + fichierACharger);
      ex.printStackTrace();
    }

    jeu = true; (thread = new Thread(this)).start();
  }

  /** 
   * Elle se charge de dessiner le decors,
   * le heros et les monstres en utilisant le double buffer.
   * @param g C'est le graphics que l'on va dessiner.
   */
  public final void paint(Graphics g) {
    // Si la variable jeu est a true 
    if(jeu) {
      db = createImage(largeur, hauteur);
      dbGraphics = db.getGraphics();
      
      level.dessineToi(dbGraphics, heros.getClef());
      // on dessine le heros de le meme contexte graphique que le decors.
      heros.dessineToi(dbGraphics);
      // on gere les interactions du heros dans le decors.
      heros.verification();

      // on dessine les monstres dans le decors.
      for(int i = 0 ; i < nbMonstre ; i++) {
	tableauDeMonstre[i].dessineToi(dbGraphics);
      }
      // Gestion des monstres dans le decors en fonction du heros.
      for(int i = 0 ; i < nbMonstre ; i++) {
  	tableauDeMonstre[i].pousuitHeros(heros);
      }
      // On recopie le double buffer.
      g.drawImage(db, 0, 0, this);
    }
    // Si la variable jeu est a faux.
    else {
      try {
	g.drawImage(Toolkit.getDefaultToolkit().getImage("./sprite/" + imageAAfficher), 0, 0, this);
      } catch(Exception ex) {
	System.err.println("CLASS FenetreJeu : echec lors de la lecture de l'image : " + imageAAfficher);
	ex.printStackTrace();
      }
    }
    touche = true;
  }

  /**
   * Pour gerer le reaffichage de la fenetre.
   *  
   * @param g C'est le graphics que l'on va dessiner.
   */
  public final void update(Graphics g) {    
    paint(g);
  }

  /**
   * Elle gere le thread qui dessine le Canvas.
   */
  public void run() {
    while(true) {
      // On redessine les elements du jeu.
      repaint();
      // pour que les actions du clavier fonctionnent sous Linux.
      requestFocus();
      try {
        //Entre chaque redessinage, on attend 500ms.
	thread.sleep(500);
      } catch(InterruptedException e) {
	System.err.println("CLASS FenetreJeu : sleep erreur !");
	e.printStackTrace();
      }
    }
  }

  // accesseurs
  /**
   * La taille minimale de la fenetre de visualisation.
   * @return une fenetre (largeur * hauteur).
   */ 
  public Dimension getMinimunDimension() {
    return getPreferredSize();
  }

  /**
   * Pour que la fenetre dessinee fasse au moins une
   * taille minimale.
   * @return Dimension La taille de la fenetre.
   */
  public Dimension getPreferredSize() {
    return new Dimension(largeur,hauteur);
  }

  /**
   * Pour acceder au heros courant.
   * @return heros retourne le heros courant.
   */
  public Heros getHeros() {
    return heros;
  }

  /**
   * Pour acceder au level courant.
   * @return level retourne le level courant.
   */
  public Level getLevel() {
    return level;
  }  

  // gestion des touches
  /**
   * Gestion des evenements sur le clavier
   * @param keyEvent touche clavier utilisee
   */
  public void keyTyped(KeyEvent keyevent) {}

  /** 
   * Gestion des evenements sur le clavier
   * @param keyEvent touche clavier utilisee
   */
  public void keyReleased(KeyEvent keyevent) {
    int code = keyevent.getKeyCode();
    if(touche) {
      switch(code) {
      case 37 : 
      case 38 : 
      case 39 : 
      case 40 : heros.deplaceToi(code) ; break;
      case 32 : heros.creuse() ; break;
      case 17 : heros.dynamite(); break;
      case 10 : jeu = true; (thread = new Thread(this)).start(); break;
      }
      touche = false;
    }
    }

  /** 
   * Gestion des evenements sur le clavier
   *  @param keyEvent touche clavier utilisee
   */
  public void keyPressed(KeyEvent keyevent) {}

}
